# ttp.tcl
#
# LEG20070912
# LEG20070827
#
# main ttp include file:
#   ttp parser and template functions
#
# This file is part of TTP the Tcl Template Parser
# TTP is (c) 2007 Georg Lehner <jorge@magma.com.ni>
# Share and use it as you like, but don't blame me

namespace eval opt {
    proc define {n c p {b {}}} {
	variable comment
	set comment($n) $c

	if [llength $b] {
	    proc $n [concat $p { } args] \
		[append b \n {eval options $args}]
	} else {
	    proc $n args [append p \n {eval options $args}]}}

    proc alias {name args} {
	foreach alias [join $args] {
	    interp alias {} ::opt::$alias {} ::opt::$name}}

    proc set {varName value} {::set ::opt::$varName $value}

    proc lappend {varName value} {::lappend ::opt::$varName $value}

    proc usage {{msg ""} {cmdline "\[opts\] args"}} {
	variable comment

	::set options [namespace eval ::opt info proc -*]
	::set t "Usage: [file tail $::argv0] $cmdline\n"
	puts stderr $t
	foreach o [lsort $options] {
	    ::set ot "  $o "
	    append ot [lrange [namespace eval ::opt info args $o] 0 end-1]
	    append ot "\t.. " $comment($o)
	    puts stderr $ot}
	puts stderr "$msg\n"}

    proc options args {
	if [llength $args] {
	    if [catch {eval $args} err] {
		::log::fatal "bad option:$args: $err!"}}}

    variable argv {}
    proc unknown {opt args} {
	switch -glob -- $opt {
	    -* {::log::fatal "bad option: $opt $args!"}
	    default {
		lappend argv $opt
		eval options $args}}}

    proc parse {args} {
	rename ::unknown ::opt::tclUnknown
	proc ::unknown {args} {eval ::opt::unknown $args}
	if [llength $args] {
	    eval options [join $args]
	} else {eval options $::argv}
	rename ::unknown {}
	rename ::opt::tclUnknown ::unknown}


    # Default commandline options for ttp
    proc defaults {} {

	define -D "define a variable" {varName value} {::set ::$varName $value}
	alias  -D --define --def

	define -I "set include path" path {set root $path}

	variable libs {}
	define -l "source in a library" file {lappend libs $file}

	define -h "usage information" {
	    usage "\nloglevels: [array names ::log::levels]" "\[options\] \[arguments\]"
	    exit 0}
	alias  -h --help

	define -d "set debuging level, default: $::log::level" value {
	    if ![info exists ::log::levels($value)] {
		fatal "wrong parameter for -d: $value" -1
	    } else {::set ::log::level $value}}

	variable output
	variable ofd stdout
	define -o "output to file instead of stdout" path {
	    if [catch {open $path w} ::opt::ofd] {
		fatal "-o $path: $::opt::ofd" -1
	    } else {set output $path}}

	variable maxErrs 200
	define -e "break after count parsing errors, default: $maxErrs" {value} {set maxErrs $value}
    }

    proc argv {} {# parse argv, exit if an error ocurred
	defaults; parse
	if $::log::rc {::opt::usage; exit $::log::rc}}
}

namespace eval ttp {
    variable state parse cmdLine {}

    proc out line {
	variable state

	switch $state {
	    comment {}
	    default {puts $::opt::ofd $line}}}

    proc parseError msg {
	upvar line line fileName fileName lineNumber lineNumber
	incr ::log::rc
	log ERROR "$fileName:$lineNumber: $msg"
	out ">>> $line"
	if {$::log::rc > $::opt::maxErrs} {fatal "to many parse errors: $::log::rc"}}

    proc doCmd {line fileName lineNumber} {
	variable cmdLine

	log DEVEL "doCmd: $line, $fileName, $lineNumber."
	# TODO: why are we called so that line must be "join"ed
	#set $line [join $line]
	if ![llength $line] return
	if {[string index $line 0] eq "\#"} return

	set script [concat $cmdLine $line]
	log DEBUG $script
	if [catch $script err] {parseError "$script:$err"}}

    proc parseLine {line fileName lineNumber} {
	variable state

	switch $state {
	    cmd {}
	    comment -
	    literal {catch {uplevel \#0 [list subst $line]} result}

	    literalstart -  commentstart -  tclstart -  cmdstart -
	    tclend -  literalend -
	    commentend {log ERROR "$fileName:$lineNumber:unexpected state: $state: $line"}

	    tcl {
		if [catch {uplevel \#0 eval $line} result] {
		    parseError $result}}
	    default {
		if [catch {uplevel \#0 [list subst $line]} result] {
		    parseError $result}}}

	# output, or not?
	switch $state {

	    literalstart {set state literal}
	    tclstart {set state tcl}
	    cmdstart {set state cmd}

	    tclend -  skipline -  literalend -
	    commentend {set state parse}

	    tclline -  literalline -
	    commentline {
		if [string length $result] {out $result}
		set state parse}

	    tcl -
	    parse {out $result}
	    literal {out $line}
	    cmd {doCmd [join $line] $fileName $lineNumber}
	    comment {}

	    default {
		parseError "internal error: $state: $result"}}}

    proc parse {fileName {initState parse} {initCmdLine {}}} {
	variable state
	variable cmdLine $initCmdLine

	if [catch {open $fileName r} fd] {
	    fatal "opening template $fileName: $fd"}

	if [llength [info procs ::preambel]] {preambel $fileName}

	set lineNumber 0
	set state $initState
	while {[gets $fd line]!=-1} {
	    parseLine $line $fileName [incr lineNumber]}

	if [llength [info procs ::postambel]] {postambel $fileName}

	skipline}

    proc skipline {} {
	variable state
	set state skipline
	return}

    # Comment
    proc -- {args} {
	variable state
	if [llength $args] {
	    switch $state {
		parse {set state commentline}}
	} else {
	    switch $state {
		parse {set state comment}
		comment {set state commentend}}}
	return}

    # literal: don't subst
    proc literal {{text {}}} {
	variable state
	if [string length $text] {
	    switch $state {
		parse {set state literalline}}
	    return $text
	} else {
	    switch $state {
		parse {set state literalstart}
		literal {set state literalend}}
	    return}}

    # tcl: eval instead of subst
    proc tcl {args} {
	variable state
	if [llength $args] {
	    switch $state {
		parse {set state tclline}}
	    catch {eval $args} result
	    return $result
	} else {
	    switch $state {
		parse {set state tclstart}
		tcl {set state tclend}}}
	return}

    # cmd: preproces lines instead of subst
    proc cmd {args} {
	variable state
	variable cmdLine
	switch $state {
	    parse {
		set cmdLine $args
		set state cmdstart}}}

    namespace export out parse skipline -- literal tcl cmd
}

namespace import ::ttp::*

# Utility commands in template files (and elsewhere)
proc stamp {} {
    set host ""
    if [info exists ::env(HOST)] {set host $::env(HOST)}
    if [info exists ::env(HOSTNAME)] {set host $::env(HOSTNAME)}
    if {$host eq ""} {catch {exec hostname} host}

    set user ""
    if [info exists ::env(USER)] {set user $::env(USER)}
    if [info exists ::env(LOGNAME)] {set user $::env(LOGNAME)}

    out "# ttp: [clock format [clock seconds]]
#   ${user}@${host}:$::env(PWD)
#   $::argv0 $::argv"
    skipline}

proc let: {name value args} {# [name] returns value
    # a [variable] is simply a procedure
    proc $name {} [list return $value]
    if [llength $args] skipline; return}

proc set: {name value args} {
    # set name $value, global name, discard args (comments)
    # a set command without return value, name is made global in the
    # calling scope
    set ::$name $value
    uplevel [subst -nocommands \
	     {if ![info exists $name] {global $name}}]
    if [llength $args] skipline; return}

# if name is not set (in global scope) set it to the given value
proc default: {name value args} {
    # if not yet set, set name to value, make name global
    if ![info exists ::$name] {
	set ::$name $value}
    uplevel [subst -nocommands \
		 {if ![info exists $name] {global $name}}]
    if [llength $args] skipline; return}

proc set? {name value} {# either $name or value
    # if variable name exists return its value
    # else return the given (default) value
    #
    # the syntax is based on 'set name' which returns $name
    if [uplevel [list info exists $name]] {uplevel set $name} \
    else {return $value}}

proc unset! {name} {# unset
    if [uplevel [list info exists $name]] {uplevel unset $name}}
